#! /usr/bin/env python
"""Base fixtures for unit tests"""

from collections import namedtuple
from datetime import datetime
from mock import Mock

import git
from git.util import IterableList
import pytest


@pytest.fixture
def mock_repo():
    """repo mock fixture"""
    repo_mock = Mock()
    repo_git_mock = Mock()
    repo_mock.attach_mock(repo_git_mock, 'git')

    remote = Mock(spec=git.Remote)
    remote.configure_mock(name="origin", url="http://example.com")
    remote_list = IterableList("name")
    remote_list.extend([remote])

    repo_mock.remotes = remote_list
    repo_mock.branches = []

    return repo_mock


@pytest.fixture
def fake_commits(count=3):
    """A few commit-like objects, to test log_diff functions"""
    Author = namedtuple("Author", ["name", "email"])
    Commit = namedtuple("Commit", ["hexsha", "message", "summary", "author",
                                   "authored_datetime"])

    commits = []
    for i in range(0, count):
        author = Author(name="Test Author", email="testauthor@example.com")
        msg = "This is a commit message (#{0})\nWith some details.".format(i)

        commit = Commit(
            hexsha="00{0}0000000000000000".format(i),
            message=msg,
            summary="This is a commit message (#{0})".format(i),
            author=author,
            authored_datetime=datetime(2018, 12, 5, 10, 36, 19)
        )
        commits.append(commit)

    return commits



#! /usr/bin/env python
"""This module acts as an interface for acting on git remotes"""

import git
from future.utils import raise_from

from git_wrapper import exceptions


class GitRemote(object):

    def __init__(self, git_repo, logger):
        """Constructor for GitRemote object

            :param repo.GitRepo git_repo: An already constructed GitRepo object to use
            :param logging.Logger logger: A pre-configured Python Logger object
        """
        self.git_repo = git_repo
        self.logger = logger

    def names(self):
        """Returns a list of remotes for a given repo

            :return list: A list of utf-8 encoded remote names
        """
        return [x.name for x in self.git_repo.repo.remotes]

    def add(self, name, url):
        """Adds a remote to the given repo

            :param str name: The name for the remote
            :param str url: The url to use for the remote
            :return bool: True if the remote was added, False otherwise
        """
        self.logger.debug("Adding remote %s (%s) to repo %s", name, url, self.git_repo.repo.working_dir)
        ret_status = False

        try:
            remote = self.git_repo.repo.create_remote(name, url)
        except git.CommandError as ex:
            self.logger.debug("Failed to create new remote %s (%s). Error: %s", name, url, ex)
            return ret_status

        try:
            remote.update()
            ret_status = True
        except git.CommandError as ex:
            self.logger.debug("Failed to update new remote %s (%s), removing it. Error: %s", name, url, ex)
            self.git_repo.repo.delete_remote(remote)

        return ret_status

    def fetch(self, remote="origin"):
        """Refresh the specified remote

           :param str remote: Remote to fetch
        """
        try:
            remote = self.git_repo.repo.remote(remote)
        except ValueError:
            msg = "Remote {remote} does not exist on repo {repo}".format(
                remote=remote,
                repo=self.git_repo.repo.working_dir
            )
            raise exceptions.ReferenceNotFoundException(msg)

        try:
            remote.fetch()
        except git.GitCommandError as ex:
            msg = (
                "Could not fetch remote {remote} ({url}). "
                "Error: {error}".format(remote=remote.name,
                                        url=remote.url,
                                        error=ex)
            )
            raise_from(exceptions.RemoteException(msg), ex)

    def fetch_all(self):
        """Refresh all the repo's remotes.

           All the remotes will be fetched even if one fails ; in this case a
           single exception containing the list of failed remotes is returned.
        """
        remotes = self.names()

        errors = []
        for remote in remotes:
            try:
                self.fetch(remote)
            except exceptions.RemoteException:
                self.logger.exception("Error fetching remote %s", remote)
                errors.append(remote)

        if errors:
            msg = "Error fetching these remotes: {0}".format(", ".join(errors))
            raise exceptions.RemoteException(msg)



#! /usr/bin/env python
"""Tests for GitRepo"""

from mock import Mock, patch, ANY
import shutil

import git
import pytest

from git_wrapper import exceptions
from git_wrapper.branch import GitBranch
from git_wrapper.commit import GitCommit
from git_wrapper.remote import GitRemote
from git_wrapper.repo import GitRepo
from git_wrapper.tag import GitTag


def test_repo(mock_repo):
    """
    GIVEN GitRepo initialized with a path and no repo object
    WHEN the object is created
    THEN a repo added
    """
    with patch('git_wrapper.repo.git') as git_mock:
        attrs = {'Repo.return_value': mock_repo}
        git_mock.configure_mock(**attrs)
        git_util = GitRepo('./')
        assert mock_repo == git_util.repo


def test_not_path_no_repo():
    """
    GIVEN GitRepo initialized with no path or repo object
    WHEN the object is created
    THEN an exception is raised
    """
    with pytest.raises(Exception):
        GitRepo('', None)


def test_git_command(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN the git property is called
    THEN a git object is returned
    """
    git_util = GitRepo('./', mock_repo)

    assert mock_repo.git is git_util.git


def test_remote_setter(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the remote setter is called
    THEN the remote is set as expected
    """
    repo = GitRepo('./', mock_repo)
    new_remote = GitRemote(git_repo=repo, logger=None)
    repo.remote = new_remote
    assert repo.remote == new_remote


def test_remote_setter_wrong_type(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the remote setter is called with the wrong type
    THEN a TypeError is raised
    """
    repo = GitRepo('./', mock_repo)
    with pytest.raises(TypeError):
        repo.remote = repo


def test_branch_setter(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the branch setter is called
    THEN the branch is set as expected
    """
    repo = GitRepo('./', mock_repo)
    new_branch = GitBranch(git_repo=repo, logger=None)
    repo.branch = new_branch
    assert repo.branch == new_branch


def test_branch_setter_wrong_type(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the branch setter is called with the wrong type
    THEN a TypeError is raised
    """
    repo = GitRepo('./', mock_repo)
    with pytest.raises(TypeError):
        repo.branch = repo


def test_commit_setter(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the commit setter is called
    THEN the commit is set as expected
    """
    repo = GitRepo('./', mock_repo)
    new_commit = GitCommit(git_repo=repo, logger=None)
    repo.commit = new_commit
    assert repo.commit == new_commit


def test_commit_setter_wrong_type(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the commit setter is called with the wrong type
    THEN a TypeError is raised
    """
    repo = GitRepo('./', mock_repo)
    with pytest.raises(TypeError):
        repo.commit = repo


def test_tag_setter(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the tag setter is called
    THEN the tag is set as expected
    """
    repo = GitRepo('./', mock_repo)
    new_tag = GitTag(git_repo=repo, logger=None)
    repo.tag = new_tag
    assert repo.tag == new_tag


def test_tag_setter_wrong_type(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN the tag setter is called with the wrong type
    THEN a TypeError is raised
    """
    repo = GitRepo('./', mock_repo)
    with pytest.raises(TypeError):
        repo.tag = repo


def test_clone():
    """
    GIVEN GitRepo without a path or repo
    WHEN clone is called with a valid clone_from URL and clone_to path
    THEN Repo.clone_from is called
    """
    with patch('git.repo.base.Repo.clone_from') as mock_clone:
        clone = GitRepo.clone('./', './testclone')
        assert mock_clone.called is True
        assert isinstance(clone, GitRepo)


def test_bare_clone():
    """
    GIVEN GitRepo without a path or repo
    WHEN clone is called with valid parameters and bare set to True
    THEN Repo.clone_from is called with bare=True
    """
    with patch('git.repo.base.Repo.clone_from') as mock_clone:
        GitRepo.clone('./', './testclone', True)
        mock_clone.assert_called_with('./', ANY, bare=True)


def test_clone_failed():
    """
    GIVEN GitRepo without a path or repo
    WHEN clone is called with a valid clone_from URL and clone_to path
    AND Repo.clone_from fails with an exception
    THEN a RepoCreationException is raised
    """
    with patch('git.repo.base.Repo.clone_from') as mock_clone:
        mock_clone.side_effect = git.GitCommandError('clone', '')
        with pytest.raises(exceptions.RepoCreationException):
            GitRepo.clone('./', './testclone')


def test_destroy_and_reclone(mock_repo, monkeypatch):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN destroy_and_reclone is called
    THEN Repo.clone_from is called
    WITH the expected remote url and local working dir
    """
    monkeypatch.setattr(shutil, 'rmtree', Mock())
    clone = GitRepo(repo=mock_repo)
    local_dir = '/tmp/8f697668fgitwrappertest'
    clone.repo.working_dir = local_dir

    with patch('git.repo.base.Repo.clone_from') as mock_clone:
        clone.destroy_and_reclone()
        assert mock_clone.called is True
        mock_clone.assert_called_with('http://example.com',
                                      local_dir,
                                      bare=False)


def test_destroy_no_path_no_repo(monkeypatch):
    """
    GIVEN GitRepo initialized with no path or repo object
    WHEN destroy_and_reclone is called
    THEN an exception is raised
    """
    monkeypatch.setattr(shutil, 'rmtree', Mock())
    with pytest.raises(Exception):
        clone = GitRepo('', None)
        clone.destroy_and_reclone()


def test_destroy_no_remotes(mock_repo, monkeypatch):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN destroy_and_reclone is called
    AND the repo does not have any remotes configured
    THEN an exception is raised
    """
    monkeypatch.setattr(shutil, 'rmtree', Mock())
    clone = GitRepo(repo=mock_repo)

    with pytest.raises(exceptions.RepoCreationException):
        clone.repo.remotes = {}
        clone.destroy_and_reclone()


def test_destroy_no_remote_named_origin(mock_repo, monkeypatch):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN destroy_and_reclone is called
    AND the repo does not have a remote named origin
    THEN Repo.clone_from is called
    WITH the remote url and local working dir from another remote
    """
    monkeypatch.setattr(shutil, 'rmtree', Mock())
    clone = GitRepo(repo=mock_repo)
    local_dir = '/tmp/8f697667fgitwrappertest'
    clone.repo.working_dir = local_dir

    remote = Mock(spec=git.Remote)
    remote.configure_mock(name="onlyremote", url="http://example.com/another")
    clone.repo.remotes = [remote]

    with patch('git.repo.base.Repo.clone_from') as mock_clone:
        clone.destroy_and_reclone()
        assert mock_clone.called is True
        mock_clone.assert_called_with('http://example.com/another',
                                      local_dir,
                                      bare=False)


def test_destroy_and_multiple_remotes(mock_repo, monkeypatch):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN destroy_and_reclone is called
    AND the repo has multiple remotes
    THEN Repo.clone_from is called
    AND create_remote is called
    """
    monkeypatch.setattr(shutil, 'rmtree', Mock())
    clone = GitRepo(repo=mock_repo)
    local_dir = '/tmp/8f697668fgitwrappertest'
    clone.repo.working_dir = local_dir

    remote = Mock(spec=git.Remote)
    remote.configure_mock(name="otherremote", url="http://example.com/another")
    clone.repo.remotes.append(remote)

    with patch('git.repo.base.Repo.clone_from') as mock_clone:
        new_repo_mock = Mock()
        mock_clone.return_value = new_repo_mock
        clone.destroy_and_reclone()
        assert mock_clone.called is True
        mock_clone.assert_called_with('http://example.com',
                                      local_dir,
                                      bare=False)
        new_repo_mock.create_remote.assert_called_with(
            "otherremote",
            "http://example.com/another"
        )


def test_destroy_and_remote_creation_fails(mock_repo, monkeypatch):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN destroy_and_reclone is called
    AND the repo has several remotes
    AND create_remote fails
    THEN a RemoteException is raised
    """
    monkeypatch.setattr(shutil, 'rmtree', Mock())
    clone = GitRepo(repo=mock_repo)
    local_dir = '/tmp/8f697668fgitwrappertest'
    clone.repo.working_dir = local_dir

    remote = Mock(spec=git.Remote)
    remote.configure_mock(name="otherremote", url="http://example.com/another")
    clone.repo.remotes.append(remote)

    with patch('git.repo.base.Repo.clone_from') as mock_clone:
        new_repo_mock = Mock()
        mock_clone.return_value = new_repo_mock
        with pytest.raises(exceptions.RemoteException):
            new_repo_mock.create_remote.side_effect = git.GitCommandError('remote', '')
            clone.destroy_and_reclone()
        assert mock_clone.called is True



import pytest

from git_wrapper import exceptions
from git_wrapper.repo import GitRepo


def test_tag(repo_root):
    repo = GitRepo(repo_root)

    # Make sure we are on master
    repo.git.checkout("master")
    head = repo.repo.head.object.hexsha

    # Test tag creation by tagging the current head
    repo.tag.create("test_tag", head)

    # The new tag and "master" references should match
    assert repo.commit.same("test_tag", "master") is True

    # Test tag deletion
    repo.tag.delete("test_tag")

    # Confirm tag doesn't exist anymore
    with pytest.raises(exceptions.ReferenceNotFoundException):
        repo.tag.delete("test_tag")



#! /usr/bin/env python
"""Tests for GitTag"""

from mock import patch

import git
import pytest

from git_wrapper import exceptions
from git_wrapper.repo import GitRepo


def test_create_tag(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.create is called with a valid name and reference
    THEN git.create_tag is called
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.tag.create("my_tag", "123456")
    repo.repo.create_tag.assert_called_with("my_tag", "123456")


def test_create_tag_with_wrong_ref(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.create is called with a name and invalid reference
    THEN a ReferenceNotFoundException is raised
    AND git.create_tag is not called
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.tag.create("my_tag", "123456")
    repo.repo.create_tag.assert_not_called()


def test_create_tag_failed(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.create is called with a valid name and reference
    AND git.create_tag fails
    THEN a TaggingException is raised
    """
    repo = GitRepo(repo=mock_repo)
    mock_repo.create_tag.side_effect = git.GitCommandError('create_tag', '')

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.TaggingException):
            repo.tag.create("my_tag", "123456")


def test_delete_tag(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.delete is called with a valid tag name
    THEN git.tag is called
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.tag.delete("my_tag")
    repo.git.tag.assert_called_with("-d", "my_tag")


def test_delete_tag_that_doesnt_exist(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.delete is called with an invalid tag name
    THEN a ReferenceNotFoundException is raised
    AND git.tag is not called
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.tag.delete("bad_tag")
    repo.git.tag.assert_not_called()


def test_delete_tag_failed(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.delete is called with a valid tag name
    AND git.tag fails
    THEN a TaggingException is raised
    """
    repo = GitRepo(repo=mock_repo)
    mock_repo.git.tag.side_effect = git.GitCommandError('delete_tag', '')

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.TaggingException):
            repo.tag.delete("my_tag")


def test_push_tag(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.push is called with a valid tag name and remote
    THEN git.push is called
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.tag.push("my_tag", "origin")
    repo.git.push.assert_called_with("origin", "my_tag")


def test_push_tag_as_dry_run(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.push is called with a valid tag name and remote
    AND dry_run is set to true
    THEN git.push is called with -n as first argument
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.tag.push("my_tag", "origin", True)
    repo.git.push.assert_called_with("-n", "origin", "my_tag")


def test_push_tag_to_bad_remote(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.push is called with a tag name and invalid remote
    THEN a ReferenceNotFoundException is raised
    AND git.push is not called
    """
    repo = GitRepo(repo=mock_repo)

    with patch("git.repo.fun.name_to_object"):
        with pytest.raises(exceptions.ReferenceNotFoundException) as exc_info:
            repo.tag.push("my_tag", "bad_remote")

    assert 'bad_remote' in str(exc_info.value)
    repo.git.push.assert_not_called()


def test_push_tag_failed(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN tag.push is called with a valid tag name and remote
    AND git.push fails
    THEN a PushException is raised
    """
    repo = GitRepo(repo=mock_repo)
    repo.git.push.side_effect = git.GitCommandError('push', '')

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.PushException):
            repo.tag.push("my_tag", "origin")



#! /usr/bin/env python
"""This module acts as an interface for acting on git commits"""

import git
from future.utils import raise_from

from git_wrapper import exceptions
from git_wrapper.utils.decorators import reference_exists


class GitCommit(object):

    def __init__(self, git_repo, logger):
        """Constructor for GitCommit object

            :param repo.GitRepo git_repo: An already constructed GitRepo object to use
            :param logging.Logger logger: A pre-configured Python Logger object
        """
        self.git_repo = git_repo
        self.logger = logger

    @reference_exists('sha')
    def describe(self, sha):
        """Return tag and commit info for a given sha

           :param str sha: The SHA1 of the commit to describe
           :return dict: A dict with tag and patch data
        """
        ret_data = {'tag': '', 'patch': ''}
        try:
            output = self.git_repo.git.describe('--all', sha).split('-g')
        except git.CommandError as ex:
            msg = "Error while running describe command on sha {sha}: {error}".format(sha=sha, error=ex)
            raise_from(exceptions.DescribeException(msg), ex)

        if output:
            tag = output[0]
            # Lightweight tags have a tag/ prefix when returned
            if tag.startswith('tag/'):
                tag = tag[4:]
            ret_data['tag'] = tag
            if len(output) > 1:
                ret_data['patch'] = output[1]
        return ret_data

    def commit(self, message, signoff=False):
        """Create a commit for changes to tracked files in the repo.
           Equivalent to `git commit -a -m <message>`.

           :param str message: The commit message
           :param bool signoff: Whether to add signed-off-by to commit message
        """
        if not message or not isinstance(message, str):
            self.logger.debug("Cannot create commit without commit message.")
            raise exceptions.CommitMessageMissingException('No commit message text provided.')

        # Add tracked files to the index
        self.git_repo.git.add(update=True)
        changes = self.git_repo.git.diff(name_only=True, staged=True)
        if not changes:
            self.logger.info("No changes to commit.")
            return False

        # Commit the changes
        self.logger.debug("Preparing to commit changes to the following files: %s", changes)
        commit = self.git_repo.git.commit(message=message, all=True, signoff=signoff)
        self.logger.info("Committed changes as commit %s", commit)

    @reference_exists('hash_')
    def revert(self, hash_, message=None):
        """Revert a specified commit.

            :param str hash_: The commit hash or reference to rebase to
            :param str message: Extra info to be included in commit message
        """
        try:
            self.git_repo.git.revert(hash_, no_edit=True)
        except git.GitCommandError as ex:
            msg = "Revert failed for hash {hash_}. Error: {error}".format(
                hash_=hash_, error=ex)
            raise_from(exceptions.RevertException(msg), ex)

        if message:
            commit = git.repo.fun.name_to_object(self.git_repo.repo, hash_)
            self.git_repo.git.commit(
                "--amend",
                "-m", 'Revert "{0}"'.format(commit.summary),
                "-m", "This reverts commit {0}.".format(commit.hexsha),
                "-m", message
            )

    @reference_exists('reference_B')
    @reference_exists('reference_A')
    def same(self, reference_A, reference_B):
        """Determine whether two references refer to the same commit.

            :param str reference_A: A commit ref (sha, tag, branch name, ...)
            :param str reference_B: A commit ref (sha, tag, branch name, ...)
            :return bool: True if the references point to the same commit,
                          False if not
        """
        commitA = git.repo.fun.name_to_object(self.git_repo.repo, reference_A)
        commitB = git.repo.fun.name_to_object(self.git_repo.repo, reference_B)

        if commitA.hexsha == commitB.hexsha:
            return True
        else:
            return False



import pytest

from git_wrapper import exceptions
from git_wrapper.repo import GitRepo


def test_rebase(repo_root, rebase_cleanup):
    repo = GitRepo(repo_root)
    branch_name = "mybranches/test_repo"
    rebase_to = "2e6c014bc296be90a7ed04d155ea7d9da2240bbc"  # Hash for 0.1.0 tag

    assert repo.repo.active_branch.name == "master"
    assert repo.repo.head.object.hexsha != rebase_to

    # Create a branch based on an old tag
    repo.repo.git.branch(branch_name, "0.0.1")

    # Rebase that branch
    repo.branch.rebase_to_hash(branch_name=branch_name, hash_=rebase_to)
    assert repo.repo.active_branch.name == branch_name
    assert repo.repo.head.object.hexsha == rebase_to


def test_abort_rebase(repo_root, rebase_cleanup):
    repo = GitRepo(repo_root)

    # Set the stage for a failed rebase
    repo.repo.git.checkout("2d88955411f2bd2162f24455f8e948ce435152c5")
    repo.repo.git.cherry_pick("5ded3c1362229c874dea3ac8d63b89b0b104c57a")
    current_head = repo.repo.head.object.hexsha

    # Fail that rebase
    assert repo.repo.is_dirty() is False
    with pytest.raises(exceptions.RebaseException):
        repo.branch.rebase_to_hash(current_head, "31777bbb03da53424c2b0eeae2504a237a4f1720")
    assert repo.repo.is_dirty() is True

    # Perform abort and ensure resulting repo status is clean
    repo.branch.abort_rebase()
    assert repo.repo.is_dirty() is False
    assert repo.repo.head.object.hexsha == current_head


def test_abort_rebase_failure(repo_root):
    repo = GitRepo(repo_root)

    with pytest.raises(exceptions.AbortException):
        repo.branch.abort_rebase()



#! /usr/bin/env python
"""This module acts as an interface for common git tasks"""

import logging
import os
import shutil

import git
from future.utils import raise_from

from git_wrapper.branch import GitBranch
from git_wrapper.commit import GitCommit
from git_wrapper import exceptions
from git_wrapper.remote import GitRemote
from git_wrapper.tag import GitTag


class GitRepo(object):
    """Provides a wrapper to interact with a git repository"""

    def __init__(self, path='', repo=None, logger=None):
        """Constructor for GitRepo object

            :param str path: Path to a git repo Default('')
            :param git.Repo repo: An already constructed git.Repo object to use Default(None)
            :param logging.Logger logger: A pre-configured Python Logger object
        """
        self.__repo = None  # Added to clear pylint warnings
        self._branch = None
        self._commit = None
        self._remote = None
        self._tag = None

        self.__setup(path, repo)
        self._setup_logger(logger)

    def _setup_logger(self, logger):
        """Set up a pre-configured logger or create a new one

            :param logging.Logger logger: A pre-configured Python Logger object
        """
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)

    def __setup(self, path, repo):
        """Sets the path and repo after performing validation

            :param str path: Path to a directory containing a git repository
            :param git.Repo repo: An already constructed git.Repo object to use
        """
        if repo is not None:
            self.repo = repo
        elif path != '':
            self.repo = git.Repo(path)
        else:
            raise Exception('No path or repo given')

    @property
    def repo(self):
        """Returns the git repo for a given path

            :return git.Repo: A reference to the internal git.Repo object
        """
        return self.__repo

    @repo.setter
    def repo(self, new_repo):
        """Allows for use of an already constructed repo

            :param git.Repo new_repo: An already constructed git.Repo object to use
        """
        self.__repo = new_repo

    @property
    def git(self):
        """Returns the git command for a given repo"""
        return self.repo.git

    @staticmethod
    def clone(clone_from, clone_to, bare=False):
        """Clone a repository.

           :param str clone_from: The url or path to clone the repo from
           :param str clone_to: The local path to clone to
           :param bool bare: Whether to create a bare repo
           :return GitRepo: Returns the newly created repo object
        """
        clone_to = os.path.realpath(os.path.expanduser(clone_to))
        logging.debug("Preparing to clone repository %s into directory %s",
                      clone_from, clone_to)

        try:
            repo = git.repo.base.Repo.clone_from(clone_from,
                                                 clone_to,
                                                 bare=bare)
        except git.GitCommandError as ex:
            msg = "Error cloning repository {repo}".format(repo=clone_from)
            raise_from(exceptions.RepoCreationException(msg), ex)

        return GitRepo(repo=repo)

    def destroy_and_reclone(self):
        """Deletes the current directory and reclone the repository."""
        # Get local path for the repo
        local_path = self.repo.working_dir

        self.logger.info("Preparing to delete and reclone repo %s", local_path)

        # Get all of the remotes info
        remotes = {}
        for r in self.repo.remotes:
            remotes[r.name] = r.url

        self.logger.debug("Remotes for %s: %s", local_path, ' '.join(list(remotes)))

        if len(remotes) == 0:
            msg = "No remotes found for repo {repo}, cannot reclone. Aborting deletion.".format(repo=local_path)
            raise exceptions.RepoCreationException(msg)

        # Select a remote for the clone, 'origin' by default
        if "origin" in remotes.keys():
            default_remote = "origin"
        else:
            default_remote = list(remotes)[0]

        self.logger.debug("Default remote for cloning set to '%s'", default_remote)

        # Delete the local repo
        self.logger.info("Deleting local repo at %s", local_path)
        shutil.rmtree(local_path, ignore_errors=True)

        # Clone it again
        repo = self.clone(remotes[default_remote], local_path)
        self.__setup(repo=repo.repo, path=local_path)

        # Recreate the remotes
        for name, url in remotes.items():
            if name != default_remote:
                try:
                    self.logger.debug("Adding remote %s", name)
                    r = self.repo.create_remote(name, url)
                except git.GitCommandError as ex:
                    msg = "Issue with recreating remote {remote}. Error: {error}".format(remote=name, error=ex)
                    raise_from(exceptions.RemoteException(msg), ex)

    @property
    def remote(self):
        """Return object to act on the repo's remotes"""
        if not self._remote:
            self._remote = GitRemote(git_repo=self, logger=self.logger)
        return self._remote

    @remote.setter
    def remote(self, new_remote):
        """Set up object to interact with Remotes

            :param git_wrapper.remote.GitRemote new_remote: An already constructed GitRemote object to use
        """
        if not isinstance(new_remote, GitRemote):
            raise TypeError("Remote must be a GitRemote object.")
        self._remote = new_remote

    @property
    def branch(self):
        """Return object to act on the repo's branches"""
        if not self._branch:
            self._branch = GitBranch(git_repo=self, logger=self.logger)
        return self._branch

    @branch.setter
    def branch(self, new_branch):
        """Set up object to interact with Branches

            :param git_wrapper.branch.GitBranch new_branch: An already constructed GitBranch object to use
        """
        if not isinstance(new_branch, GitBranch):
            raise TypeError("Branch must be a GitBranch object.")
        self._branch = new_branch

    @property
    def commit(self):
        """Return object to act on the repo's commits"""
        if not self._commit:
            self._commit = GitCommit(git_repo=self, logger=self.logger)
        return self._commit

    @commit.setter
    def commit(self, new_commit):
        """Set up object to interact with Commits

            :param git_wrapper.commit.GitCommit new_commit: An already constructed GitCommit object to use
        """
        if not isinstance(new_commit, GitCommit):
            raise TypeError("Commit must be a GitCommit object.")
        self._commit = new_commit

    @property
    def tag(self):
        """Return object to act on the repo's tags"""
        if not self._tag:
            self._tag = GitTag(git_repo=self, logger=self.logger)
        return self._tag

    @tag.setter
    def tag(self, new_tag):
        """Set up object to interact with Tags

            :param git_wrapper.tag.GitTag new_tag: Pre-constructed GitTag object
        """
        if not isinstance(new_tag, GitTag):
            raise TypeError("Tag must be a GitTag object.")
        self._tag = new_tag



import git
import pytest

from git_wrapper import exceptions
from git_wrapper.repo import GitRepo


def test_apply_diff(repo_root, datadir):
    repo = GitRepo(repo_root)
    test_branch = "test_apply_diff"

    # Create a diff file
    diff_path = (datadir / "test.diff")

    # Create a branch from a commit the diff will apply cleanly to
    repo.git.branch(test_branch, "90946e854499ee371c22f6a492fd0f889ae2394f")
    assert repo.repo.active_branch.name == 'master'

    # Apply it
    repo.branch.apply_diff(test_branch, diff_path, "Test commit message", True)

    # Check latest commit
    assert repo.repo.active_branch.name == test_branch
    message = repo.repo.head.object.message
    assert "Test commit message" in message
    assert "Signed-off-by" in message


def test_apply_patch(repo_root, patch_cleanup, datadir):
    repo = GitRepo(repo_root)
    test_branch = "test_apply_patch"

    # Create patch file (based on git format-patch)
    patch_path = (datadir / "test.patch")

    # Create a branch from a commit the patch will apply cleanly to
    repo.git.branch(test_branch, "0.1.0")
    assert repo.repo.active_branch.name == 'master'

    # Apply & check
    repo.branch.apply_patch(test_branch, patch_path)
    assert repo.repo.active_branch.name == test_branch
    assert "Test patch" in repo.repo.head.object.message


def test_abort(repo_root, patch_cleanup, datadir):
    repo = GitRepo(repo_root)
    test_branch = "test_abort"

    # Create a bad patch file
    patch_path = (datadir / "test-bad.patch")

    # Create a branch from a commit
    repo.git.branch(test_branch, "0.1.0")
    assert repo.repo.active_branch.name == 'master'

    # Apply & check for the failure
    with pytest.raises(exceptions.ChangeNotAppliedException):
        repo.branch.apply_patch(test_branch, patch_path)
    assert repo.repo.active_branch.name == test_branch

    # Revert
    repo.branch.abort_patch_apply()
    assert "Test patch" not in repo.repo.head.object.message


def test_log_diff(repo_root):
    repo = GitRepo(repo_root)

    log_diff = repo.branch.short_log_diff(
        "155b270a26c5c6cc5e90281fb1a741b293364caa",
        "964aaddddadfbf646fea6a4549e4cb3661a35712"
    )

    assert len(log_diff) == 2
    assert log_diff[0] == "964aadd Drop py34"
    assert log_diff[1] == "5ded3c1 Add basic logging in more places"


def test_log_diff_no_results(repo_root):
    repo = GitRepo(repo_root)

    log_diff = repo.branch.log_diff("0.1.0", "0.1.0")

    assert len(log_diff) == 0


def test_log_diff_wrong_hash(repo_root):
    repo = GitRepo(repo_root)

    with pytest.raises(exceptions.ReferenceNotFoundException):
        repo.branch.log_diff("0.1.0", "123456789z")

    with pytest.raises(exceptions.ReferenceNotFoundException):
        repo.branch.log_diff("123456789z", "0.1.0")


def test_reset(repo_root):
    repo = GitRepo(repo_root)
    branch_name = "test_reset"

    # Exercise repo refresh
    repo.remote.fetch("origin")

    # Save the current reference to origin/master
    reset_to_commit = git.repo.fun.name_to_object(repo.repo, "origin/master")

    # Create a new branch based on an old commit
    repo.git.branch(branch_name, "0.0.1")

    # Ensure branch head is different from the one we saved
    branch_commit = repo.repo.branches[branch_name].commit
    assert branch_commit.hexsha != reset_to_commit.hexsha

    # Reset the branch to origin/master
    repo.branch.hard_reset(
        refresh=False,  # Avoid race condition if something new merged
        branch=branch_name,
        remote="origin",
        remote_branch="master"
    )

    # Ensure the new head matches the origin/master we saved
    branch_commit = repo.repo.branches[branch_name].commit
    assert branch_commit.hexsha == reset_to_commit.hexsha


def test_create_branch(repo_root):
    repo = GitRepo(repo_root)
    branch_name = "test_create"
    tag_0_0_1_hexsha = "631b3a35723a038c01669e1933571693a166db81"
    tag_0_1_0_hexsha = "2e6c014bc296be90a7ed04d155ea7d9da2240bbc"

    assert branch_name not in repo.repo.branches

    # Create the new branch
    repo.branch.create(branch_name, "0.0.1")
    assert branch_name in repo.repo.branches
    assert repo.repo.branches[branch_name].commit.hexsha == tag_0_0_1_hexsha

    # Branch already exists - do nothing
    repo.branch.create(branch_name, "0.1.0")
    assert branch_name in repo.repo.branches
    assert repo.repo.branches[branch_name].commit.hexsha == tag_0_0_1_hexsha

    # Branch already exists - reset it
    repo.branch.create(branch_name, "0.1.0", True)
    assert branch_name in repo.repo.branches
    assert repo.repo.branches[branch_name].commit.hexsha == tag_0_1_0_hexsha


def test_create_and_checkout_branch(repo_root):
    repo = GitRepo(repo_root)
    branch_name = "test_create"

    assert repo.repo.active_branch.name == 'master'

    # Create and check out the new branch
    repo.branch.create(branch_name, "0.0.1", checkout=True)
    assert repo.repo.active_branch.name == branch_name

    repo.repo.heads.master.checkout()
    assert repo.repo.active_branch.name == 'master'

    # Branch already exists - reset it and don't check it out
    repo.branch.create(branch_name, "0.1.0", True, checkout=False)
    assert repo.repo.active_branch.name == 'master'

    # Branch already exists - reset it and check it out
    repo.branch.create(branch_name, "0.0.1", True, checkout=True)
    assert repo.repo.active_branch.name == branch_name


def test_remote_contains(repo_root, patch_cleanup, datadir):
    repo = GitRepo(repo_root)
    remote_branch = "origin/master"

    # 1. Check a known commit
    assert repo.branch.remote_contains(
        remote_branch, "fc88bcb3158187ba9566dad896e3c688d8bc5109"
    ) is True

    # 2. Confirm new commit doesn't exist on the remote
    test_branch = "test_contains"
    patch_path = (datadir / "test.patch")
    repo.git.branch(test_branch, "0.1.0")  # For patch to apply cleanly
    repo.branch.apply_patch(test_branch, patch_path)

    assert repo.branch.remote_contains(
        remote_branch, repo.repo.head.object.hexsha
    ) is False



class GitWrapperException(Exception):
    """Superclass for this library's exceptions"""
    pass


class DirtyRepositoryException(GitWrapperException):
    """Repository workspace is dirty."""
    pass


class ReferenceNotFoundException(GitWrapperException):
    """Reference (commit, tag, branch, ...) doesn't exist."""
    pass


class CheckoutException(GitWrapperException):
    """Error occurred while switching branch."""
    pass


class RebaseException(GitWrapperException):
    """Error occurred during a rebase."""
    pass


class AbortException(GitWrapperException):
    """Error occurred while attempting to abort a command."""
    pass


class RevertException(GitWrapperException):
    """Error occurred while attempting to perform a revert."""
    pass


class DescribeException(GitWrapperException):
    """Error occurred while running the describe command."""
    pass


class CommitMessageMissingException(GitWrapperException):
    """Cannot create a commit without a commit message."""
    pass


class ChangeNotAppliedException(GitWrapperException):
    """Error occurred while applying a patch or diff onto a repo."""
    pass


class FileDoesntExistException(GitWrapperException):
    """File doesn't exist."""
    pass


class RepoCreationException(GitWrapperException):
    """Error occurred while creating or cloning a repo."""
    pass


class RemoteException(GitWrapperException):
    """Error occurred while doing an operation on a remote."""
    pass


class ResetException(GitWrapperException):
    """Error occurred while resetting."""
    pass


class TaggingException(GitWrapperException):
    """Error occurred while performing operation on tag."""
    pass


class PushException(GitWrapperException):
    """Error occurred while pushing to remote."""
    pass



#! /usr/bin/env python
"""This module acts as an interface for acting on git tags"""

import git
from future.utils import raise_from

from git_wrapper import exceptions
from git_wrapper.utils.decorators import reference_exists


class GitTag(object):

    def __init__(self, git_repo, logger):
        """Constructor for GitTag object

            :param repo.GitRepo git_repo: An already constructed GitRepo object
            :param logging.Logger logger: A pre-configured Python Logger object
        """
        self.git_repo = git_repo
        self.logger = logger

    @reference_exists('reference')
    def create(self, name, reference):
        """Create a new tag to the target reference.

           :param str name: New tag's name
           :param str reference: What the tag should point to
        """
        try:
            self.git_repo.repo.create_tag(name, reference)
        except git.GitCommandError as ex:
            msg = "Error creating tag {name} on {ref}. Error: {error}".format(
                name=name, ref=reference, error=ex
            )
            raise_from(exceptions.TaggingException(msg), ex)

    @reference_exists('name')
    def delete(self, name):
        """Delete tag from local repository

           :param str name: Tag name to delete
        """
        try:
            self.git_repo.git.tag("-d", name)
        except git.GitCommandError as ex:
            msg = "Error deleting tag {name}. Error: {error}".format(
                name=name, error=ex
            )
            raise_from(exceptions.TaggingException(msg), ex)

    @reference_exists('name')
    def push(self, name, remote, dry_run=False):
        """Push specified tag to specified remote

           :param str name: Tag name
           :param str remote: Remote to push the tag to
           :param bool dry_run: Whether to run the commands in dry-run mode
        """
        if remote not in self.git_repo.remote.names():
            msg = "No remote named {0}".format(remote)
            raise exceptions.ReferenceNotFoundException(msg)

        msg = (
            "Error pushing tag {name} (dry-run: {dry_run}) to remote "
            "{remote}.".format(name=name, remote=remote, dry_run=dry_run)
        )

        try:
            if dry_run:
                self.git_repo.git.push("-n", remote, name)
            else:
                self.git_repo.git.push(remote, name)
        except git.GitCommandError as ex:
            raise_from(exceptions.PushException(msg), ex)



#! /usr/bin/env python
"""Tests for GitCommit"""

from mock import Mock, patch

import git
import pytest

from git_wrapper.repo import GitRepo
from git_wrapper import exceptions


def test_on_head_only_all_new(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN on_head_only method is called with no upstream equivalent changes
    THEN a dictionary is returned containing two sha1's and commits
    """
    repo = GitRepo('./', mock_repo)
    lines = '+ sha1 commit1\n+ sha2 commit2\n+ sha3 commit3'
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    expected = {'sha1': 'commit1', 'sha2': 'commit2', 'sha3': 'commit3'}
    assert expected == repo.branch.cherry_on_head_only('upstream', 'HEAD')


def test_on_head_only_with_mixed(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN on_head_only method is called with a mix of
         upstream equivalent and not equivalent changes
    THEN a dictionary is returned containing two sha1's and commits
    """
    repo = GitRepo('./', mock_repo)
    lines = '+ sha1 commit1\n- sha2 commit2\n+ sha3 commit3'
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    expected = {'sha1': 'commit1', 'sha3': 'commit3'}
    assert expected == repo.branch.cherry_on_head_only('upstream', 'HEAD')


def test_on_head_only_no_new(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN on_head_only method is called with a only upstream equivalent changes
    THEN an empty dictionary is returned
    """
    repo = GitRepo('./', mock_repo)
    lines = '- sha1 commit1\n- sha2 commit2\n- sha3 commit3'
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    assert {} == repo.branch.cherry_on_head_only('upstream', 'HEAD')


def test_on_head_only_empty(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN on_head_only is called with no changes
    THEN an empty dictionary is returned
    """
    repo = GitRepo('./', mock_repo)
    lines = ''
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    assert {} == repo.branch.cherry_on_head_only('upstream', 'HEAD')


def test_all_equivalent_changes(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN equivalent is called with only equivalent upstream/downstream changes.
    THEN a dictionary is returned with all changes
    """
    repo = GitRepo('./', mock_repo)
    lines = '- sha1 commit1\n- sha2 commit2\n- sha3 commit3'
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    expected = {'sha1': 'commit1', 'sha2': 'commit2', 'sha3': 'commit3'}
    assert expected == repo.branch.cherry_equivalent('upstream', 'HEAD')


def test_equivalent_mixed_changes(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN equivalent is called with mix equivalent and HEAD changes.
    THEN a dictionary is returned with only the equivalent changes.
    """
    repo = GitRepo('./', mock_repo)
    lines = '+ sha1 commit1\n- sha2 commit2\n+ sha3 commit3'
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    expected = {'sha2': 'commit2'}
    assert expected == repo.branch.cherry_equivalent('upstream', 'HEAD')


def test_equivalent_downstream_only(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN equivalent is called with mix HEAD only changes.
    THEN an empty dictionary is returned.
    """
    repo = GitRepo('./', mock_repo)
    lines = '+ sha1 commit1\n+ sha2 commit2\n+ sha3 commit3'
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    assert {} == repo.branch.cherry_equivalent('upstream', 'HEAD')


def test_equivalent_no_changes(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN equivalent is called with no changes.
    THEN an empty dictionary is returned.
    """
    repo = GitRepo('./', mock_repo)
    lines = ''
    attrs = {'cherry.return_value': lines}
    mock_repo.git.configure_mock(**attrs)
    assert {} == repo.branch.cherry_equivalent('upstream', 'HEAD')


def test_rebase(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.rebase_to_hash is called with a valid branch name and a valid hash
    THEN git.checkout called
    AND git.rebase called
    """
    mock_repo.is_dirty.return_value = False
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.branch.rebase_to_hash('test', '12345')

    assert repo.repo.git.checkout.called is True
    assert repo.repo.git.rebase.called is True


def test_rebase_dirty_repo(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.rebase_to_hash is called on a dirty repository
    THEN a DirtyRepositoryException is raised
    """
    mock_repo.is_dirty.return_value = True
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.DirtyRepositoryException):
            repo.branch.rebase_to_hash('test', '12345')
    assert mock_repo.is_dirty.called is True


def test_rebase_branch_not_found(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.rebase_to_hash is called with an invalid branch name
    THEN a ReferenceNotFoundException is raised
    AND the exception message contains branch
    """
    mock_repo.is_dirty.return_value = False
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        with pytest.raises(exceptions.ReferenceNotFoundException) as exc_info:
            mock_name_to_object.side_effect = git.exc.BadName()
            repo.branch.rebase_to_hash('doesNotExist', '12345')
    assert 'branch' in str(exc_info.value)


def test_rebase_hash_not_found(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.rebase_to_hash is called with a valid branch name and an invalid hash
    THEN a ReferenceNotFoundException is raised
    AND the exception message contains hash
    """
    mock_repo.is_dirty.return_value = False
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        with pytest.raises(exceptions.ReferenceNotFoundException) as exc_info:
            # First name_to_object call is to check the branch, let it succeed
            def side_effect(mock, ref):
                if ref != "branchA":
                    raise git.exc.BadName
            mock_name_to_object.side_effect = side_effect
            repo.branch.rebase_to_hash('branchA', '12345')
    assert 'hash' in str(exc_info.value)


def test_rebase_error_during_checkout(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.rebase_to_hash is called with a valid branch name and a valid hash
    AND checkout fails with an exception
    THEN a CheckoutException is raised
    """
    mock_repo.is_dirty.return_value = False
    mock_repo.git.checkout.side_effect = git.GitCommandError('checkout', '')
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.CheckoutException):
            repo.branch.rebase_to_hash('branchA', '12345')


def test_rebase_error_during_rebase(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.rebase_to_hash is called with a valid branch name and a valid hash
    AND rebase fails with an exception
    THEN a RebaseException is raised
    """
    mock_repo.is_dirty.return_value = False
    mock_repo.git.rebase.side_effect = git.GitCommandError('rebase', '')
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.RebaseException):
            repo.branch.rebase_to_hash('branchA', '12345')


def test_abort_rebase(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.abort_rebase is called
    THEN git.rebase called
    """
    repo = GitRepo('./', mock_repo)

    repo.branch.abort_rebase()
    assert repo.repo.git.rebase.called is True


def test_abort_rebase_error(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN abort_rebase is called
    AND the abort fails with an exception
    THEN an AbortException is raised
    """
    mock_repo.git.rebase.side_effect = git.GitCommandError('rebase', '')
    repo = GitRepo('./', mock_repo)

    with pytest.raises(exceptions.AbortException):
        repo.branch.abort_rebase()


def test_apply_patch(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_patch is called with a valid branch_name and valid path
    THEN git.am is called
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.branch.apply_patch('test_branch', './requirements.txt')
    assert repo.git.am.called is True


def test_apply_patch_wrong_branch_name(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_patch is called with an invalid branch_name and valid path
    THEN ReferenceNotFoundExceptionRaised
    AND git.am not called
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.branch.apply_patch('invalid_branch', './requirements.txt')
    assert repo.git.am.called is False


def test_apply_patch_not_a_file(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_patch is called with a valid branch_name and invalid path
    THEN FileDoesntExistException raised
    AND git.am not called
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.FileDoesntExistException):
            repo.branch.apply_patch('test_branch', './git_wrapper')
    assert repo.git.am.called is False


def test_apply_patch_checkout_error(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_patch is called with a valid branch name and a valid path
    AND checkout fails with an exception
    THEN a CheckoutException is raised
    AND git.am not called
    """
    mock_repo.git.checkout.side_effect = git.GitCommandError('checkout', '')
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.CheckoutException):
            repo.branch.apply_patch('test_branch', './requirements.txt')
    assert repo.git.am.called is False


def test_apply_patch_apply_error(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_patch is called with a valid branch name and a valid path
    AND git.am fails with an exception
    THEN a ChangeNotAppliedException is raised
    """
    mock_repo.git.am.side_effect = git.GitCommandError('am', '')
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.ChangeNotAppliedException):
            repo.branch.apply_patch('test_branch', './requirements.txt')


def test_apply_diff(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called with a valid branch_name and valid diff_path and valid message
    THEN index.commit is called
    """
    mock_repo.is_dirty.return_value = False
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.branch.apply_diff('test_branch', './requirements.txt', 'message', True)
    assert repo.git.apply.called is True
    assert repo.git.commit.called is True


def test_apply_diff_on_invalid_branch(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called with an invalid branch_name and valid path
    THEN ReferenceNotFoundExceptionRaised
    AND git.apply not called
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.branch.apply_diff('invalid_branch', './requirements.txt', 'message')
    assert repo.git.apply.called is False


def test_apply_diff_on_dirty_workspace(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called on a dirty repository
    THEN a DirtyRepositoryException is raised
    AND git.apply not called
    """
    mock_repo.is_dirty.return_value = True
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.DirtyRepositoryException):
            repo.branch.apply_diff('test_branch', './requirements.txt', 'message')
    assert mock_repo.is_dirty.called is True
    assert repo.git.apply.called is False


def test_apply_diff_no_commit_message(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called with valid branch_name, valid diff_path and invalid message
    THEN CommitMessageMissingException raised
    AND index.commit not called
    """
    mock_repo.is_dirty.return_value = False
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.CommitMessageMissingException):
            repo.branch.apply_diff('test_branch', './requirements.txt', '')
    assert repo.git.commit.called is False


def test_apply_diff_not_a_file(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called with valid parameters
    THEN FileDoesntExistException raised
    AND git.apply not called
    """
    mock_repo.is_dirty.return_value = False
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.FileDoesntExistException):
            repo.branch.apply_diff('test_branch', 'doesntexist.txt', 'message')
    assert repo.git.apply.called is False


def test_apply_diff_checkout_error(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called with valid parameters
    AND checkout fails with an exception
    THEN a CheckoutException is raised
    AND index.commit not called
    """
    mock_repo.is_dirty.return_value = False
    mock_repo.git.checkout.side_effect = git.GitCommandError('checkout', '')
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.CheckoutException):
            repo.branch.apply_diff('invalid_branch', './requirements.txt', 'my message')
    assert repo.git.commit.called is False


def test_apply_diff_apply_fails(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called with a valid branch_name and valid diff_path and valid message
    AND git.apply fails with an exception
    THEN an ChangeNotAppliedException is raised
    """
    mock_repo.is_dirty.return_value = False
    mock_repo.git.apply.side_effect = git.GitCommandError('apply', '')
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.ChangeNotAppliedException):
            repo.branch.apply_diff('test_branch', './requirements.txt', 'message')
    assert repo.git.commit.called is False


def test_apply_diff_apply_nothing_to_commit(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN apply_diff is called with a valid branch_name and valid diff_path and valid message
    WHEN commit is called with a valid message
    AND there are no diff changes
    THEN git.apply called
    AND index.commit not called
    """
    mock_repo.is_dirty.return_value = False
    repo = GitRepo('./', mock_repo)
    repo.git.diff.return_value = []

    with patch('git.repo.fun.name_to_object'):
        repo.branch.apply_diff('test_branch', './requirements.txt', 'message')
    assert repo.git.apply.called is True
    assert repo.git.commit.called is False


def test_abort_patch_apply(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN abort_patch_apply is called
    THEN git.am called
    """
    repo = GitRepo('./', mock_repo)

    repo.branch.abort_patch_apply()
    assert repo.git.am.called is True


def test_abort_patch_apply_error(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN abort_patch_apply is called
    AND the abort_patch_apply fails with an exception
    THEN an Abort_Patch_ApplyException is raised
    """
    mock_repo.git.am.side_effect = git.GitCommandError('abort_patch_apply', '')
    repo = GitRepo('./', mock_repo)

    with pytest.raises(exceptions.AbortException):
        repo.branch.abort_patch_apply()


def test_reverse_diff(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN reverse_diff is called with a valid diff_path
    THEN git.am called
    """
    repo = GitRepo('./', mock_repo)

    repo.branch.reverse_diff('./requirements.txt')
    assert repo.git.apply.called is True


def test_reverse_diff_diff_file_doesnt_exist(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN reverse_diff is called with and invalid diff_path
    THEN FileDoesntExistException raised
    AND git.apply not called
    """
    repo = GitRepo('./', mock_repo)

    with pytest.raises(exceptions.FileDoesntExistException):
        repo.branch.reverse_diff('./thisdoesntexist')
    assert repo.git.apply.called is False


def test_reverse_diff_error(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN reverse_diff is called with a valid diff_path
    AND the reverse_diff fails with an exception
    THEN an RevertException is raised
    """
    mock_repo.git.apply.side_effect = git.GitCommandError('apply', '')
    repo = GitRepo('./', mock_repo)

    with pytest.raises(exceptions.RevertException):
        repo.branch.reverse_diff('./requirements.txt')


def test_log_diff(mock_repo, fake_commits):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN log_diff is called with two valid hashes
    THEN a list of log entries is returned
    """
    mock_repo.iter_commits.return_value = fake_commits
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        log_diff = repo.branch.log_diff('12345', '54321')

    assert len(log_diff) == 3
    assert log_diff[2] == (
        "commit 0020000000000000000\n"
        "Author: Test Author <testauthor@example.com>\n"
        "Date: Wed Dec 05 10:36:19 2018 \n\n"
        "This is a commit message (#2)\nWith some details."
    )


def test_short_log_diff(mock_repo, fake_commits):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN short_log_diff is called with two valid hashes
    THEN a list of log entries is returned
    """
    mock_repo.iter_commits.return_value = fake_commits
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        log_diff = repo.branch.short_log_diff('12345', '54321')

    assert len(log_diff) == 3
    assert log_diff == ["0000000 This is a commit message (#0)",
                        "0010000 This is a commit message (#1)",
                        "0020000 This is a commit message (#2)"]


def test_log_diff_with_pattern(mock_repo, fake_commits):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN log_diff is called with two valid hashes and a pattern
    THEN a list of log entries is returned
    """
    mock_repo.iter_commits.return_value = fake_commits
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        log_diff = repo.branch.log_diff('12345', '54321',
                                        pattern="$hash $author")

    assert log_diff == [
        "0000000000000000000 Test Author <testauthor@example.com>",
        "0010000000000000000 Test Author <testauthor@example.com>",
        "0020000000000000000 Test Author <testauthor@example.com>"
    ]


def test_log_diff_no_results(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN log_diff is called with two valid hashes
    AND there are no results
    THEN an empty list is returned
    """
    mock_repo.iter_commits.return_value = []
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        log_diff = repo.branch.log_diff('12345', '12345')

    assert log_diff == []


def test_log_diff_invalid_hash(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN log_diff is called with a invalid hash
    THEN a ReferenceNotFoundException is raised
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        with pytest.raises(exceptions.ReferenceNotFoundException):
            mock_name_to_object.side_effect = git.exc.BadName()
            repo.branch.log_diff('doesNotExist', '12345')

    assert mock_repo.iter_commits.called is False


def test_reset(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset is called
    THEN repo.head.reset is called
    """
    mock_remote = Mock()
    mock_repo.remote.return_value = mock_remote

    repo = GitRepo(repo=mock_repo)
    with patch('git.repo.fun.name_to_object'):
        repo.branch.hard_reset()

    assert mock_remote.fetch.called is True  # Sync is called
    assert mock_repo.head.reset.called is True  # Reset is called


def test_reset_remote_reference_not_found(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset is called
    AND the remote + branch reference doesn't exist
    THEN ReferenceNotFoundException is raised
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.branch.hard_reset(refresh=False, remote="doesntExist")

    assert mock_repo.head.reset.called is False


def test_reset_checkout_failure(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset is called
    AND git.checkout fails
    THEN CheckoutException is raised
    """
    mock_repo.git.checkout.side_effect = git.GitCommandError('checkout', '')
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.CheckoutException):
            repo.branch.hard_reset(refresh=False)

    assert mock_repo.head.reset.called is False


def test_reset_reset_failure(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset is called
    AND git.reset fails
    THEN ResetException is raised
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        mock_repo.head.reset.side_effect = git.GitCommandError('reset', '')
        with pytest.raises(exceptions.ResetException):
            repo.branch.hard_reset(refresh=False)


def test_reset_to_ref_with_checkout(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset is called with checkout
    THEN repo.head.reset is called
    AND repo.checkout is called once
    """
    repo = GitRepo(repo=mock_repo)
    with patch('git.repo.fun.name_to_object'):
        repo.branch.hard_reset_to_ref("main", "origin/main", checkout=True)

    assert mock_repo.head.reset.called is True
    assert mock_repo.git.checkout.call_count == 1


def test_reset_to_ref_detached_head_with_checkout(mock_repo, monkeypatch):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset is called with checkout
    AND the current HEAD is detached
    THEN repo.head.reset is called
    AND repo.checkout is called once
    """
    class MockRef:
        @property
        def name(self):
            # Detached heads don't have a name
            raise TypeError

    repo = GitRepo(repo=mock_repo)
    with patch('git.repo.fun.name_to_object'):
        mock_repo.head.ref = MockRef()
        repo.branch.hard_reset_to_ref("main", "origin/main", checkout=True)

    assert mock_repo.head.reset.called is True
    assert mock_repo.git.checkout.call_count == 1


def test_reset_to_ref_without_checkout(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset_to_ref is called with checkout False
    THEN repo.head.reset is called
    AND repo.checkout is called twice to return to the original state
    """
    repo = GitRepo(repo=mock_repo)
    with patch('git.repo.fun.name_to_object'):
        repo.branch.hard_reset_to_ref("main", "origin/main", checkout=False)

    assert mock_repo.head.reset.called is True
    assert mock_repo.git.checkout.call_count == 2


def test_reset_to_ref_without_checkout_fails(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN reset_to_ref is called with checkout False
    AND switching back fails
    THEN checkoutException is raised
    """
    mock_repo.git.checkout.side_effect = [None, git.GitCommandError('checkout', '')]

    repo = GitRepo(repo=mock_repo)
    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.CheckoutException):
            repo.branch.hard_reset_to_ref("main", "origin/main", checkout=False)

    assert mock_repo.head.reset.called is True
    assert mock_repo.git.checkout.call_count == 2


def test_local_branch_exists(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.exists is called with a valid branch and None remote
    THEN True is returned
    """
    repo = GitRepo(repo=mock_repo)
    mock_repo.branches = ["master", "test"]

    assert repo.branch.exists("test") is True


def test_local_branch_doesnt_exist(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.exists is called with an invalid branch and None remote
    THEN False is returned
    """
    repo = GitRepo(repo=mock_repo)
    mock_repo.branches = ["master", "test"]

    assert repo.branch.exists("another-test") is False


def test_branch_exists_with_invalid_remote(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.exists is called with a valid branch and invalid remote
    THEN a RemoteException is raised
    """
    repo = GitRepo(repo=mock_repo)

    with pytest.raises(exceptions.RemoteException):
        assert repo.branch.exists("another", "doesntexist")


def test_remote_branch_exists(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.exists is called with a valid branch and valid remote
    THEN True is returned
    """
    repo = GitRepo(repo=mock_repo)

    remote = Mock(spec=git.Remote)
    remote.configure_mock(name="testremote", refs=["testbranch"])
    mock_repo.remotes.extend([remote])

    assert repo.branch.exists("testbranch", "testremote") is True


def test_remote_branch_doesnt_exists(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.exists is called with an invalid branch and valid remote
    THEN True is returned
    """
    repo = GitRepo(repo=mock_repo)

    remote = Mock(spec=git.Remote)
    remote.configure_mock(name="testremote", refs=[])
    mock_repo.remotes.extend([remote])

    assert repo.branch.exists("testbranch", "testremote") is False


def test_create_branch(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.create is called with a valid name and start_ref
    THEN git.branch is called
    AND git.checkout is not called
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert repo.branch.create("test", "123456") is True
    repo.git.branch.assert_called_with("test", "123456")
    repo.git.checkout.assert_not_called()


def test_create_and_checkout_branch(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.create is called with valid parameters and checkout is True
    THEN git.branch is called
    AND git.checkout is called
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert repo.branch.create("test", "123456", checkout=True) is True
    repo.git.branch.assert_called_with("test", "123456")
    repo.git.checkout.assert_called()


def test_create_branch_with_bad_start_ref(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.create is called with a valid name and invalid start_ref
    THEN a ReferenceNotFoundException is raised
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            assert repo.branch.create("test", "badref")


def test_create_branch_already_exists(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.create is called with a valid name and start_ref
    AND the branch already exists
    THEN git.branch is not called
    """
    repo = GitRepo(repo=mock_repo)
    mock_repo.branches = ["test", "master"]

    with patch('git.repo.fun.name_to_object'):
        repo.branch.create("test", "123456")
    assert repo.git.branch.called is False
    assert repo.git.checkout.called is False


def test_create_branch_already_exists_and_check_it_out(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.create is called with valid params and checkout is True
    AND the branch already exists
    THEN git.branch is not called
    AND git.checkout is called
    """
    repo = GitRepo(repo=mock_repo)
    mock_repo.branches = ["test", "master"]

    with patch('git.repo.fun.name_to_object'):
        repo.branch.create("test", "123456", checkout=True)
    assert repo.git.branch.called is False
    assert repo.git.checkout.called is True


def test_create_branch_already_exists_and_reset_it(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.create is called with a valid name and start_ref
    AND the branch already exists and reset_if_exists is True
    THEN hard_reset_to_ref is called
    """
    repo = GitRepo(repo=mock_repo)
    mock_repo.branches = ["test", "master"]

    mock_hard_reset = Mock()
    repo.branch.hard_reset_to_ref = mock_hard_reset

    with patch('git.repo.fun.name_to_object'):
        repo.branch.create("test", "123456", True)
    assert mock_hard_reset.called is True


def test_remote_contains_branch_not_found(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.remote_contains is called with an invalid branch name
    THEN a ReferenceNotFoundException is raised
    AND the exception message contains branch
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        with pytest.raises(exceptions.ReferenceNotFoundException) as exc_info:
            mock_name_to_object.side_effect = git.exc.BadName()
            repo.branch.remote_contains('doesNotExist', '12345')
    assert 'branch' in str(exc_info.value)


def test_remote_contains_commit_not_found(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN branch.remote_contains is called with an invalid commit hash
    THEN a ReferenceNotFoundException is raised
    AND the exception message contains hash
    """
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        with pytest.raises(exceptions.ReferenceNotFoundException) as exc_info:
            # First name_to_object call is to check the branch, let it succeed
            def side_effect(mock, ref):
                if ref != "origin/mybranch":
                    raise git.exc.BadName
            mock_name_to_object.side_effect = side_effect
            repo.branch.remote_contains('origin/mybranch', 'doesNotExist')
    assert 'hash' in str(exc_info.value)


def test_remote_contains_with_commit_present(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.remote_contains is called with a valid branch and hash
    AND git_repo.git.branch returns data
    THEN branch.remote_contains returns True
    """
    remote_branch = "origin/mybranch"
    mock_repo.git.branch.return_value = remote_branch
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert repo.branch.remote_contains(remote_branch, '12345') is True


def test_remote_contains_with_commit_absent(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN branch.remote_contains is called with a valid branch and hash
    AND git_repo.git.branch returns empty string
    THEN branch.remote_contains returns True
    """
    mock_repo.git.branch.return_value = ""
    repo = GitRepo(repo=mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert repo.branch.remote_contains("origin/mybranch", '12345') is False



import shutil

import git
import pytest


CLONE_REPO_ROOT = "/tests/cloning_tests"
CLONE_REPO_URL = "https://github.com/release-depot/git_wrapper"
REPO_ROOT = "/tests/git_wrapper"  # Local copy


@pytest.fixture(scope="function")
def repo_root():
    # Check the starting point is sane
    repo = git.Repo(REPO_ROOT)
    origin = repo.remotes.origin

    repo.heads.master.checkout()
    assert repo.head.object == origin.refs.master.object
    assert repo.is_dirty() is False

    yield REPO_ROOT

    # Reset current changes and master branch to original state
    repo.head.reset(index=True, working_tree=True)
    repo.heads.master.checkout()
    repo.head.reset(origin.refs.master, index=True, working_tree=True)

    # Remove extra local branches
    for b in repo.branches:
        if b.name != 'master':
            repo.git.branch("-D", b.name)


@pytest.fixture(scope="function")
def clone_repo_root():
    # We use a different repo root for cloning tests, because some of
    # the tests are destructive and we don't want to inadvertently
    # delete the test copy of the repo that contains the patch being tested
    yield CLONE_REPO_ROOT
    shutil.rmtree(CLONE_REPO_ROOT, ignore_errors=True)


@pytest.fixture
def clone_repo_url():
    yield CLONE_REPO_URL


@pytest.fixture(scope="function")
def patch_cleanup():
    # Avoid state leaking into other tests in case of early failure
    yield
    repo = git.Repo(REPO_ROOT)
    try:
        repo.git.am('--abort')
    except git.GitCommandError:
        pass


@pytest.fixture(scope="function")
def rebase_cleanup():
    # Avoid state leaking into other tests in case of early failure
    yield
    repo = git.Repo(REPO_ROOT)
    try:
        repo.git.rebase('--abort')
    except git.GitCommandError:
        pass



#! /usr/bin/env python
"""Tests for GitRemote"""

from mock import Mock

import git
import pytest

from git_wrapper import exceptions
from git_wrapper.repo import GitRepo


def remote_generator(names):
    """Generates objects to be used with git.Repo.remotes call"""
    ret_data = []
    for name in names:
        obj = type('', (), {})()
        obj.name = name
        ret_data.append(obj)

    return ret_data


def test_get_remotes_returns_list(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN remote.names is called
    THEN a list of remote names is returned
    """
    expected = ['a', 'b', 'c']
    attrs = {'remotes': remote_generator(expected)}
    mock_repo.configure_mock(**attrs)

    git_util = GitRepo('./', mock_repo)

    assert expected == git_util.remote.names()


def test_add_remote_adds(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN remote.add is called with a name and url
    THEN a TRUE status is returned
    WITH update called
    """
    remote_mock = Mock()
    update_mock = Mock()
    remote_mock.attach_mock(update_mock, 'update')
    mock_repo.create_remote.return_value = remote_mock
    git_util = GitRepo('./', mock_repo)

    assert git_util.remote.add('rdo', 'http://rdoproject.org') is True
    assert update_mock.called is True


def test_add_remote_adds_fails(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN remote.add is called with a name and url
    AND the remote create fails with an exception
    THEN a False status is returned
    """
    mock_repo.create_remote.side_effect = git.CommandError('create')
    git_util = GitRepo('./', mock_repo)

    assert git_util.remote.add('rdo', 'http://rdoproject.org') is False


def test_add_remote_update_fails(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN remote.add is called with a name and url
    AND the remote update fails with an exception
    THEN a False status is returned
    WITH delete_remote called
    """
    remote_mock = Mock()
    delete_mock = Mock()
    update_mock = Mock(side_effect=git.CommandError('update'))
    remote_mock.attach_mock(update_mock, 'update')

    mock_repo.attach_mock(delete_mock, 'delete_remote')
    mock_repo.create_remote.return_value = remote_mock

    git_util = GitRepo('./', mock_repo)

    assert git_util.remote.add('rdo', 'http://rdoproject.org') is False
    assert update_mock.called is True
    delete_mock.assert_called_once_with(remote_mock)


def test_fetch(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN remote.fetch is called
    THEN repo.remote is called
    AND fetch is called
    """
    mock_remote = Mock()
    mock_repo.remote.return_value = mock_remote

    repo = GitRepo(repo=mock_repo)
    repo.remote.fetch()
    assert mock_repo.remote.called is True
    assert mock_remote.fetch.called is True


def test_fetch_remote_doesnt_exist(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN remote.fetch is called with a remote that doesn't exist
    THEN ReferenceNotFoundException is raised
    """
    mock_repo.remote.side_effect = ValueError

    repo = GitRepo(repo=mock_repo)
    with pytest.raises(exceptions.ReferenceNotFoundException):
        repo.remote.fetch("doesntExist")

    mock_repo.remote.assert_called_with("doesntExist")


def test_fetch_with_fetch_error(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN remote.fetch is called
    AND fetch raises an error
    THEN RemoteException is raised
    """
    mock_remote = Mock()
    mock_remote.fetch.side_effect = git.GitCommandError('fetch', '')
    mock_repo.remote.return_value = mock_remote

    repo = GitRepo(repo=mock_repo)
    with pytest.raises(exceptions.RemoteException):
        repo.remote.fetch("origin")


def test_fetch_all(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN remote.fetch_all is called
    AND there are 2 remotes
    THEN fetch is called twice
    """
    mock_remoteA, mock_remoteB = Mock(), Mock()
    mock_remotes = {"origin": mock_remoteA, "otherremote": mock_remoteB}
    mock_repo.remote = lambda r: mock_remotes[r]

    repo = GitRepo(repo=mock_repo)
    repo.remote.names = Mock(return_value=["origin", "otherremote"])

    repo.remote.fetch_all()
    assert mock_remoteA.fetch.called is True
    assert mock_remoteB.fetch.called is True


def test_fetch_all_with_errors(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN remote.fetch_all is called
    AND one remote fails
    THEN the other remotes still fetch
    AND an exception is raised
    """
    mock_remoteA, mock_remoteB, mock_remoteC = Mock(), Mock(), Mock()
    mock_remoteA.fetch.side_effect = git.GitCommandError("fetch", "")
    mock_remotes = {"origin": mock_remoteA,
                    "a_remote": mock_remoteB,
                    "other": mock_remoteC}
    mock_repo.remote = lambda r: mock_remotes[r]

    repo = GitRepo(repo=mock_repo)
    repo.remote.names = Mock(return_value=["origin", "a_remote", "other"])

    with pytest.raises(exceptions.RemoteException) as exc_info:
        repo.remote.fetch_all()

    assert 'origin' in str(exc_info.value)
    assert mock_remoteA.fetch.called is True
    assert mock_remoteB.fetch.called is True
    assert mock_remoteC.fetch.called is True



import os

import pytest

from git_wrapper.repo import GitRepo
from git_wrapper import exceptions


def test_clone(clone_repo_root, clone_repo_url):
    repo_root = clone_repo_root
    repo_url = clone_repo_url

    # Create a new clone
    clone = GitRepo.clone(repo_url, repo_root, bare=True)

    # Ensure repo has expected tags, some commits
    assert hasattr(clone, "repo") is True

    commit = clone.repo.commit("f8f7abc4ce87db051f9998c5d4dd153695e35675")
    assert commit is not None

    tag = clone.repo.commit("0.1.0")
    assert tag.hexsha == "2e6c014bc296be90a7ed04d155ea7d9da2240bbc"

    assert clone.repo.bare is True


def test_clone_from_filesystem(clone_repo_root, clone_repo_url, repo_root):
    new_repo_root = clone_repo_root
    repo_url = "file://%s" % repo_root

    # Create a new clone
    clone = GitRepo.clone(repo_url, new_repo_root)

    # Ensure repo has expected tags, some commits
    assert hasattr(clone, "repo") is True
    clone.repo.commit("f8f7abc4ce87db051f9998c5d4dd153695e35675")
    tag = clone.repo.commit("0.1.0")
    assert tag.hexsha == "2e6c014bc296be90a7ed04d155ea7d9da2240bbc"
    assert clone.repo.bare is False


def test_clone_failed(clone_repo_root, clone_repo_url):
    repo_root = clone_repo_root
    repo_url = "http://localhost"

    with pytest.raises(exceptions.RepoCreationException):
        GitRepo.clone(repo_url, repo_root)


def test_destroy_and_reclone(clone_repo_root, clone_repo_url):
    repo_root = clone_repo_root
    repo_url = clone_repo_url

    # Start by creating a new repo to destroy
    GitRepo.clone(repo_url, repo_root)

    # Get a timestamp for a random file
    f_path = os.path.join(repo_root, 'requirements.txt')
    orig_timestamp = os.path.getmtime(f_path)

    # Run the reclone action
    clone = GitRepo(repo_root)
    clone.destroy_and_reclone()

    # Ensure timestamp changed
    new_timestamp = os.path.getmtime(f_path)
    assert new_timestamp != orig_timestamp



#! /usr/bin/env python
"""Decorator utilities"""

import inspect

import git
from future.utils import raise_from
import wrapt

from git_wrapper import exceptions


def reference_exists(ref):
    @wrapt.decorator
    def wrapper(func, instance, args, kwargs):
        all_args = inspect.getcallargs(func, *args, **kwargs)
        try:
            repo = all_args['self'].git_repo.repo
            git.repo.fun.name_to_object(repo, all_args[ref])
        except git.exc.BadName as ex:
            msg = "Could not find %s %s." % (ref, all_args[ref])
            raise_from(exceptions.ReferenceNotFoundException(msg), ex)
        return func(*args, **kwargs)
    return wrapper



#! /usr/bin/env python
"""This module acts as an interface for acting on git branches"""

import re
import os
from string import Template

import git
from future.utils import raise_from

from git_wrapper import exceptions
from git_wrapper.utils.decorators import reference_exists


class GitBranch(object):

    def __init__(self, git_repo, logger):
        """Constructor for GitBranch object

            :param repo.GitRepo git_repo: An already constructed GitRepo object to use
            :param logging.Logger logger: A pre-configured Python Logger object
        """
        self.git_repo = git_repo
        self.logger = logger

    def _expand_file_path(self, path):
        """Expand a given path into an absolute path and check for presence.

           :param str path: Path
        """
        full_path = os.path.realpath(os.path.expanduser(path))
        if not os.path.isfile(full_path):
            raise exceptions.FileDoesntExistException('{path} is not a file.'.format(path=full_path))
        return full_path

    def _run_cherry(self, upstream, head, regex):
        """Run the git cherry command and return lines in a dict.

           :param str upstream: Branch name
           :param str head: Branch name
           :param str regex: Regular expression to run on the cherry result
        """
        args = ['-v', upstream, head]
        ret_data = {}
        for line in self.git_repo.git.cherry(*args).split('\n'):
            match = regex.match(line)
            if match is not None:
                ret_data[match.group(1)] = match.group(2)
        return ret_data

    @reference_exists("start_ref")
    def create(self, name, start_ref, reset_if_exists=False, checkout=False):
        """Create a local branch based on start_ref.

           If the branch exists, do nothing or hard reset it if reset_if_exists
           is set.

           :param str name: New branch's name
           :param str start_ref: Reference (branch, commit, tag, ...) to use as
                                 a starting point.
           :param bool reset_if_exists: Whether to hard reset the branch to
                                        start_ref if the branch already exists.
           :param bool checkout: Whether to checkout the new branch
           :return: True if a new branch was created, None otherwise
        """
        if not self.exists(name):
            self.git_repo.git.branch(name, start_ref)
            if checkout:
                self.git_repo.git.checkout(name)
            return True

        if self.exists(name) and reset_if_exists:
            self.hard_reset_to_ref(name, start_ref, checkout)

        if checkout:
            self.git_repo.git.checkout(name)

    def exists(self, name, remote=None):
        """Checks if a branch exists locally or on the specified remote.

           :param str name: Name of the branch to find
           :param str remote: Remote name to check for the branch, or None
                              if local
        """
        if not remote:
            if name in self.git_repo.repo.branches:
                return True
            else:
                return False
        else:
            if remote not in self.git_repo.remote.names():
                raise exceptions.RemoteException(
                    "Remote {0} does not exist.".format(remote)
                )
            if name in self.git_repo.repo.remotes[remote].refs:
                return True
            else:
                return False

    def cherry_on_head_only(self, upstream, head):
        """Get new patches between upstream and head.

           :param str upstream: Branch name
           :param str head: Branch name
        """
        self.logger.debug("Get new patches between upstream (%s) and head (%s)", upstream, head)
        head_only_regex = re.compile(r'^\+\s(.*?)\s(.*)')
        return self._run_cherry(upstream, head, head_only_regex)

    def cherry_equivalent(self, upstream, head):
        """Get patches that are in both upstream and head.

           :param str upstream: Branch name
           :param str head: Branch name
        """
        self.logger.debug("Get patches that are in both upstream (%s) and head (%s)", upstream, head)
        equivalent_regex = re.compile(r'^\-\s(.*?)\s(.*)')
        return self._run_cherry(upstream, head, equivalent_regex)

    @reference_exists("branch_name")
    @reference_exists("hash_")
    def rebase_to_hash(self, branch_name, hash_):
        """Perform a rebase from a specific reference to another.

           :param str branch_name: The name of the branch to rebase on
           :param str hash_: The commit hash or reference to rebase to
        """
        self.logger.debug(
            "Rebasing branch %s to hash %s. Repo currently at commit %s.",
            branch_name, hash_, self.git_repo.repo.head.commit
        )

        if self.git_repo.repo.is_dirty():
            msg = ("Repository {0} is dirty. Please clean workspace "
                   "before proceeding.".format(self.git_repo.repo.working_dir))
            raise exceptions.DirtyRepositoryException(msg)

        # Checkout
        try:
            self.git_repo.git.checkout(branch_name)
        except git.GitCommandError as ex:
            msg = "Could not checkout branch {name}. Error: {error}".format(name=branch_name, error=ex)
            raise_from(exceptions.CheckoutException(msg), ex)

        # Rebase
        try:
            self.git_repo.git.rebase(hash_)
        except git.GitCommandError as ex:
            msg = "Could not rebase hash {hash_} onto branch {name}. Error: {error}".format(hash_=hash_, name=branch_name, error=ex)
            raise_from(exceptions.RebaseException(msg), ex)

        self.logger.debug("Successfully rebased branch %s to %s", branch_name, hash_)

    def abort_rebase(self):
        """Aborts a rebase."""
        try:
            self.git_repo.git.rebase('--abort')
        except git.GitCommandError as ex:
            msg = "Rebase abort command failed. Error: {0}".format(ex)
            raise_from(exceptions.AbortException(msg), ex)

    @reference_exists('branch_name')
    def apply_patch(self, branch_name, path):
        """Apply a git patch file on top of the specified branch.

           :param str branch_name: The name of the branch or reference to apply the patch to
           :param str path: Path to a git-formatted patch file (cf. git format-patch)
        """
        # Expand file (also needed for git-am) and check it exists
        full_path = self._expand_file_path(path)

        # Checkout
        try:
            self.git_repo.git.checkout(branch_name)
        except git.GitCommandError as ex:
            msg = "Could not checkout branch {name}. Error: {error}".format(name=branch_name, error=ex)
            raise_from(exceptions.CheckoutException(msg), ex)

        # Apply the patch file
        try:
            self.git_repo.git.am(full_path)
        except git.GitCommandError as ex:
            msg = "Could not apply patch {path} on branch {name}. Error: {error}".format(path=full_path, name=branch_name, error=ex)
            raise_from(exceptions.ChangeNotAppliedException(msg), ex)

    @reference_exists('branch_name')
    def apply_diff(self, branch_name, diff_path, message, signoff=False):
        """Apply a diff on top of the specified branch.

           :param str branch_name: The name of the branch or reference to apply the diff to
           :param str diff_path: Path to the diff file
           :param str message: Commit message
           :param bool signoff: Whether to add signed-off-by to commit message
        """
        # Ensure we don't commit more than we mean to
        if self.git_repo.repo.is_dirty():
            msg = ("Repository {repo} contains uncommitted changes. Please clean workspace "
                   "before proceeding.".format(repo=self.git_repo.repo.working_dir))
            raise exceptions.DirtyRepositoryException(msg)

        # Check diff file exists
        full_path = self._expand_file_path(diff_path)

        # Checkout
        try:
            self.git_repo.git.checkout(branch_name)
        except git.GitCommandError as ex:
            msg = "Could not checkout branch {name}. Error: {error}".format(name=branch_name, error=ex)
            raise_from(exceptions.CheckoutException(msg), ex)

        # Apply the diff
        try:
            self.git_repo.git.apply(full_path)
        except git.GitCommandError as ex:
            msg = "Could not apply diff {path} on branch {name}. Error: {error}".format(path=full_path, name=branch_name, error=ex)
            raise_from(exceptions.ChangeNotAppliedException(msg), ex)

        # Commit
        self.git_repo.commit.commit(message, signoff)

    def abort_patch_apply(self):
        """Abort applying a patch (git am)."""
        try:
            self.git_repo.git.am('--abort')
        except git.GitCommandError as ex:
            msg = "Failed to abort git am operation. Error: {0}".format(ex)
            raise_from(exceptions.AbortException(msg), ex)

    def reverse_diff(self, diff_path):
        """Reverse a diff that was applied to the workspace.

           :param str diff_path: Path to the diff file
        """
        full_path = os.path.expanduser(diff_path)
        if not os.path.isfile(full_path):
            raise exceptions.FileDoesntExistException('{path} is not a file.'.format(path=full_path))

        try:
            self.git_repo.git.apply(full_path, reverse=True)
        except git.GitCommandError as ex:
            msg = "Reversing diff failed. Error: {0}".format(ex)
            raise_from(exceptions.RevertException(msg), ex)

    @reference_exists('hash_from')
    @reference_exists('hash_to')
    def log_diff(self, hash_from, hash_to, pattern="$full_message"):
        """Return a list of strings for log entries between two hashes.

           Any of the following placeholders may be used in the pattern:
             * $hash The full commit hash
             * $short_hash The short commit hash, similar to --abbrev-commit
             * $message The commit message
             * $summary First line of the commit message
             * $full_message Complete commit info with hash, author, message.
               Similar to default "git log" ouput
             * $author Commit author
             * $date Date the commit was authored

           :param str hash_from: A commit hash
           :param str hash_to: A commit hash
           :param str pattern: Formatter containing any of the placeholders above
        """
        range_ = "{0}..{1}".format(hash_from, hash_to)
        commits = self.git_repo.repo.iter_commits(range_)
        if not commits:
            return []

        # Prepare patterns
        author_tpl = Template("$name <$email>")
        full_message_tpl = Template(
            "commit $hash\nAuthor: $author\nDate: $date\n\n$message"
        )

        # Parse the user-provided pattern
        log = []
        line_tpl = Template(pattern)
        for c in commits:
            c_author = author_tpl.safe_substitute(name=c.author.name,
                                                  email=c.author.email)
            c_date = c.authored_datetime.strftime("%a %b %d %H:%M:%S %Y %z")

            placeholders = {"short_hash": c.hexsha[0:7],
                            "hash": c.hexsha,
                            "message": c.message,
                            "summary": c.summary,
                            "author": c_author,
                            "date": c_date}

            full_message = full_message_tpl.safe_substitute(placeholders)
            placeholders["full_message"] = full_message

            line = line_tpl.safe_substitute(placeholders)
            log.append(line)

        return log

    def short_log_diff(self, hash_from, hash_to):
        """Return a list of strings for log entries between two hashes.

           Log entries will be returned in the "<short_hash> <summary>" format.

           :param str hash_from: A commit hash
           :param str hash_to: A commit hash
        """
        return self.log_diff(hash_from, hash_to, "$short_hash $summary")

    def hard_reset(self, branch="master", remote="origin",
                   remote_branch="master", refresh=True):
        """Perform a hard reset of a local branch to a remote branch.

           :param str branch: Local branch to reset
           :param str remote: Remote use as base for the reset
           :param str remote_branch: Remote branch to reset to
           :param bool refresh: Whether to refresh the remote before resetting
        """
        if refresh:
            self.git_repo.remote.fetch(remote)

        remote_ref = "{0}/{1}".format(remote, remote_branch)
        self.hard_reset_to_ref(branch, remote_ref)

    def hard_reset_to_ref(self, branch, ref, checkout=True):
        """Perform a hard reset of a local branch to any reference.

           :param str branch: Local branch to reset
           :param str ref: Reference (commit, tag, ...) to reset to
           :param bool checkout: Whether to checkout the new branch
        """
        # Ensure the reference maps to a commit
        try:
            commit = git.repo.fun.name_to_object(self.git_repo.repo, ref)
        except git.exc.BadName as ex:
            msg = "Could not find reference {0}.".format(ref)
            raise_from(exceptions.ReferenceNotFoundException(msg), ex)

        try:
            # Preserve the reference name if there is one
            orig = self.git_repo.repo.head.ref.name
        except TypeError:
            # Detached head
            orig = self.git_repo.repo.head.commit.hexsha

        # Switch to the branch
        try:
            self.git_repo.git.checkout(branch)
        except git.GitCommandError as ex:
            msg = (
                "Could not checkout branch {branch}. Error: {error}".format(
                    branch=branch, error=ex)
            )
            raise_from(exceptions.CheckoutException(msg), ex)

        # Reset --hard to that reference
        try:
            self.git_repo.repo.head.reset(commit=commit,
                                          index=True,
                                          working_tree=True)
        except git.GitCommandError as ex:
            msg = (
                "Error resetting branch {branch} to {ref}. "
                "Error: {error}".format(ref=ref, branch=branch, error=ex)
            )
            raise_from(exceptions.ResetException(msg), ex)

        # Return to the original head if required
        if not checkout:
            try:
                self.git_repo.git.checkout(orig)
            except git.GitCommandError as ex:
                msg = (
                    "Could not checkout {orig}. Error: {error}".format(
                        orig=orig, error=ex)
                )
                raise_from(exceptions.CheckoutException(msg), ex)

    @reference_exists('remote_branch')
    @reference_exists('hash_')
    def remote_contains(self, remote_branch, hash_):
        """Check if a commit hash is present on a remote branch

           :param str remote_branch: Remote branch to check
           :param str hash_: Commit hash to check if present
        """
        # When used with a specific branch name, this command will
        # return either an empty string if the commit isn't present, or
        # the branch name provided
        result = self.git_repo.git.branch(
            "-r", "--contains", hash_, remote_branch
        )
        if result:
            return True
        else:
            return False



#! /usr/bin/env python
"""Tests for GitCommit"""

from mock import Mock, patch

import git
import pytest

from git_wrapper.repo import GitRepo
from git_wrapper import exceptions


def test_describe_tag_and_patch(mock_repo):
    """
    GIVEN GitRepo is initialized with a path and repo
    WHEN commit.describe is called with a tag and patch value
    THEN a dictionary with tag and patch keys is returned
    """
    expected = {'tag': '1.0.0', 'patch': '12345'}
    attrs = {'describe.return_value': '1.0.0-g12345'}
    mock_repo.git.configure_mock(**attrs)

    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert expected == repo.commit.describe('12345')


def test_describe_tag_only(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.describe is called a tag value only
    THEN a dictionary with a tag and empty patch is returned
    """
    expected = {'tag': '1.0.0', 'patch': ''}
    attrs = {'describe.return_value': '1.0.0'}
    mock_repo.git.configure_mock(**attrs)

    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert expected == repo.commit.describe('12345')


def test_describe_empty(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.describe is called with a bad hash
    THEN a dictionary with an empty tag and patch is returned
    """
    expected = {'tag': '', 'patch': ''}
    attrs = {'describe.return_value': ''}
    mock_repo.git.configure_mock(**attrs)

    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert expected == repo.commit.describe('12345')


def test_describe_sha_doesnt_exist(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.describe is called with a non-existent hash
    THEN a ReferenceNotFoundException is raised
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.commit.describe('doesntexist')


def test_describe_sha_with_describe_failure(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN git.describe fails
    THEN a DescribeException is raised
    """
    repo = GitRepo('./', mock_repo)
    repo.git.describe.side_effect = git.CommandError('describe')

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.DescribeException):
            repo.commit.describe('12345')


def test_describe_with_lightweight_tags(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.describe is called with a good hash for a lightweight tag
    THEN a dictionary with a tag and empty patch is returned
    AND the tag/ prefix is stripped from the tag
    """
    expected = {'tag': '1.0.0-lw', 'patch': ''}
    attrs = {'describe.return_value': 'tag/1.0.0-lw'}
    mock_repo.git.configure_mock(**attrs)

    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        assert expected == repo.commit.describe('12345')


def test_commit(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.commit is called with a valid message
    THEN index.commit called
    """
    repo = GitRepo('./', mock_repo)
    repo.git.diff.return_value = '+- change'

    repo.commit.commit("My commit message")
    assert repo.git.commit.called is True


def test_commit_no_changes(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.commit is called with a valid message
    AND there are no diff changes
    THEN git.diff called
    AND index.commit not called
    """
    repo = GitRepo('./', mock_repo)
    repo.git.diff.return_value = []

    repo.commit.commit('my commit message')
    assert repo.git.diff.called is True
    assert repo.git.commit.called is False


def test_commit_no_message(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.commit is called with no valid message
    THEN CommitMessageMissingException raised
    """
    repo = GitRepo('./', mock_repo)

    with pytest.raises(exceptions.CommitMessageMissingException):
        repo.commit.commit('')


def test_commit_message_is_not_a_string(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.commit is called with no valid message
    THEN CommitMessageMissingException raised
    """
    repo = GitRepo('./', mock_repo)

    with pytest.raises(exceptions.CommitMessageMissingException):
        repo.commit.commit(True)


def test_revert(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.revert is called with a valid hash_ and no message
    THEN git.revert called
    AND git.commit is not called
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.commit.revert('123456')
    assert repo.git.revert.called is True
    # commit only called when there's a message explicitly included
    assert repo.git.commit.called is False


def test_revert_hash_not_found(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.revert is called
    AND the hash_ doesn't exist
    THEN a ReferenceNotFoundException is raised
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.commit.revert('123456')


def test_revert_error(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.revert is called
    AND the revert fails with an exception
    THEN a RevertException is raised
    """
    mock_repo.git.revert.side_effect = git.GitCommandError('revert', '')
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        with pytest.raises(exceptions.RevertException):
            repo.commit.revert('123456')


def test_revert_with_message(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.revert is called with a valid hash_ and a message
    THEN git.revert is called
    AND git.commit is called
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object'):
        repo.commit.revert('123456', "My message")
    assert repo.git.revert.called is True
    assert repo.git.commit.called is True


def test_same(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.same is called with valid references
    AND the references are to the same commit
    THEN True is returned
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.return_value = Mock(hexsha='abcd1')
        assert repo.commit.same('a_tag', 'a_commit') is True


def test_not_same(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.same is called with valid references
    AND the references are not to the same commit
    THEN False is returned
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = [Mock(hexsha='abcd1'),
                                           Mock(hexsha='zzzz2'),
                                           Mock(hexsha='abcd1'),
                                           Mock(hexsha='zzzz2')]
        assert repo.commit.same('a_tag', 'a_commit') is False


def test_same_with_invalid_ref(mock_repo):
    """
    GIVEN GitRepo initialized with a path and repo
    WHEN commit.same is called with an invalid reference
    THEN a ReferenceNotFoundException is raised
    """
    repo = GitRepo('./', mock_repo)

    with patch('git.repo.fun.name_to_object') as mock_name_to_object:
        mock_name_to_object.side_effect = git.exc.BadName()
        with pytest.raises(exceptions.ReferenceNotFoundException):
            repo.commit.same('bad_ref', 'a_tag')



import pytest

from git_wrapper import exceptions
from git_wrapper.repo import GitRepo


def test_describe(repo_root):
    repo = GitRepo(repo_root)

    with pytest.raises(exceptions.ReferenceNotFoundException):
        repo.commit.describe("doesntExist")


def test_revert(repo_root):
    repo = GitRepo(repo_root)
    repo.git.checkout("0.1.0")
    assert "history" in repo.repo.head.object.message

    repo.commit.revert("bf58876c4786fa652432f5902f8b9aef733c2f0a")

    message = repo.repo.head.object.message
    assert "Revert" in message
    assert "history" not in message


def test_revert_with_message(repo_root):
    repo = GitRepo(repo_root)
    repo.git.checkout("0.1.0")
    assert "history" in repo.repo.head.object.message

    repo.commit.revert("bf58876c4786fa652432f5902f8b9aef733c2f0a",
                       message="An explanation for the revert.")

    message = repo.repo.head.object.message
    assert "Revert" in message
    assert "This reverts commit bf5887" in message
    assert "An explanation for the revert" in message


def test_same(repo_root):
    repo = GitRepo(repo_root)

    assert repo.commit.same("0.0.1", "631b3a3") is True
    assert repo.commit.same("0.0.1", "00fed08") is False
